"""Run subprocess.
"""
import os
import sys
import platform
import subprocess

if sys.version_info > (3, 3):
    CommandNotFoundError = FileNotFoundError
else:
    CommandNotFoundError = OSError


class BadConfigError(Exception):
    """raised to halt on fatal conditions on the way to run.
    """
    pass


def get_environ(environ, defaults, ve_path):
    """Make an environment to run with.
    """
    # Copy the parent environment, add in defaults from .vexrc.
    env = environ.copy()
    env.update(defaults)

    # Leaving in existing PYTHONHOME can cause some errors
    if 'PYTHONHOME' in env:
        del env['PYTHONHOME']

    # Now we have to adjust PATH to find scripts for the virtualenv...
    # PATH being unset/empty is OK, but ve_path must be set
    # or there is nothing for us to do here and it's bad.
    if not ve_path:
        raise BadConfigError('ve_path must be set')
    if platform.system() == 'Windows':
        ve_bin = os.path.join(ve_path, 'Scripts')
    else:
        ve_bin = os.path.join(ve_path, 'bin')

    # I don't expect this to fail, but I'd rather be slightly paranoid and fail
    # early before putting a nonexistent path on PATH.
    if not os.path.exists(ve_bin):
        raise BadConfigError('ve_bin %r does not exist' % ve_bin)

    # If user is currently in a virtualenv, DON'T just prepend
    # to its path (vex foo; echo $PATH -> " /foo/bin:/bar/bin")
    # but don't incur this cost unless we're already in one.
    # activate handles this by running 'deactivate' first, we don't
    # have that so we have to use other ways.
    # This would not be necessary and things would be simpler if vex
    # did not have to interoperate with a ubiquitous existing tool.
    # virtualenv doesn't...
    current_ve = env.get('VIRTUAL_ENV', '')
    system_path = environ.get('PATH', '')
    segments = system_path.split(os.pathsep)
    if current_ve:
        # Since activate doesn't export _OLD_VIRTUAL_PATH, we are going to
        # manually remove the virtualenv's bin.
        # A virtualenv's bin should not normally be on PATH except
        # via activate or similar, so I'm OK with this solution.
        current_ve_bin = os.path.join(current_ve, 'bin')
        segments.remove(current_ve_bin)
    segments.insert(0, ve_bin)
    env['PATH'] = os.pathsep.join(segments)
    env['VIRTUAL_ENV'] = ve_path
    return env


def run(command, env, cwd):
    """Run the given command.
    """
    assert command
    if cwd:
        assert os.path.exists(cwd)
    try:
        process = subprocess.Popen(command, env=env, cwd=cwd)
        process.wait()
    except CommandNotFoundError as error:
        if error.errno != 2:
            raise
        return None
    return process.returncode



"""
This is not needed to use vex.
It just lets us provide a convenient mechanism for people
with popular shells to set up autocompletion.
"""
import os
import re

try:
    FileNotFoundError
except NameError:
    FileNotFoundError = IOError
    # (OSError, IOError)


NOT_SCARY = re.compile(br'[~]?(?:[/]+[\w _,.][\w _\-,.]+)*\Z')


def scary_path(path):
    """Whitelist the WORKON_HOME strings we're willing to substitute in
    to strings that we provide for user's shell to evaluate.

    If it smells at all bad, return True.
    """
    if not path:
        return True
    assert isinstance(path, bytes)
    return not NOT_SCARY.match(path)


def shell_config_for(shell, vexrc, environ):
    """return completion config for the named shell.
    """
    here = os.path.dirname(os.path.abspath(__file__))
    path = os.path.join(here, 'shell_configs', shell)
    try:
        with open(path, 'rb') as inp:
            data = inp.read()
    except FileNotFoundError as error:
        if error.errno != 2:
            raise
        return b''
    ve_base = vexrc.get_ve_base(environ).encode('ascii')
    if ve_base and not scary_path(ve_base):
        assert os.path.exists(ve_base)
        data = data.replace(b'$WORKON_HOME', ve_base)
    return data



"""Main command-line entry-point and any code tightly coupled to it.
"""
import sys
import os
import argparse
from vex import config
from vex.run import get_environ, run


class InvalidArgument(Exception):
    """Raised by anything under main() to propagate errors to user.
    """
    def __init__(self, message):
        self.message = message
        Exception.__init__(self, message)


class NoVirtualenvName(InvalidArgument):
    """No virtualenv name was given (insufficient arguments).
    """
    pass


class NoVirtualenvsDirectory(InvalidArgument):
    """There is no directory to find named virtualenvs in.
    """
    pass


class OtherShell(InvalidArgument):
    """The given argument to --shell-config is not recognized.
    """
    pass


class UnknownArguments(InvalidArgument):
    """Unknown arguments were given on the command line.

    This is a byproduct of having to use parse_known_args.
    """
    pass


class InvalidVexrc(InvalidArgument):
    """config file specified or required but absent or unparseable.
    """
    pass


class InvalidVirtualenv(InvalidArgument):
    """No usable virtualenv was found.
    """
    pass


class InvalidCommand(InvalidArgument):
    """No runnable command was found.
    """
    pass


class InvalidCwd(InvalidArgument):
    """cwd specified or required but unusable.
    """
    pass


def make_arg_parser():
    """Return a standard ArgumentParser object.
    """
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawTextHelpFormatter,
        usage="vex [OPTIONS] VIRTUALENV_NAME COMMAND_TO_RUN ...",
    )

    parser.add_argument(
        "--path",
        metavar="DIR",
        help="absolute path to virtualenv to use",
        action="store"
    )
    parser.add_argument(
        '--cwd',
        metavar="DIR",
        action="store",
        default='.',
        help="path to run command in (default: '.' aka $PWD)",
    )
    parser.add_argument(
        "--config",
        metavar="FILE",
        default=None,
        action="store",
        help="path to config file to read (default: '~/.vexrc')"
    )
    parser.add_argument(
        '--shell-config',
        metavar="SHELL",
        dest="shell_to_configure",
        action="store",
        default=None,
        help="print optional config for the specified shell"
    )
    parser.add_argument(
        "rest",
        nargs=argparse.REMAINDER,
        help=argparse.SUPPRESS)

    return parser


def get_options(argv):
    """Called to parse the given list as command-line arguments.

    :returns:
        an options object as returned by argparse.
    """
    arg_parser = make_arg_parser()
    options, unknown = arg_parser.parse_known_args(argv)
    if unknown:
        arg_parser.print_help()
        raise UnknownArguments("unknown args: {0!r}".format(unknown))
    options.print_help = arg_parser.print_help
    return options


def get_vexrc(options, environ):
    """Get a representation of the contents of the config file.

    :returns:
            a Vexrc instance.
    """
    # Complain if user specified nonexistent file with --config.
    # But we don't want to complain just because ~/.vexrc doesn't exist.
    if options.config and not os.path.exists(options.config):
        raise InvalidVexrc("nonexistent config: {0!r}".format(options.config))
    filename = options.config or os.path.expanduser('~/.vexrc')
    vexrc = config.Vexrc.from_file(filename, environ)
    return vexrc


def handle_shell_config(options, vexrc, environ):
    """Carry out the logic of the --shell-config option.
    """
    from vex import shell_config
    data = shell_config.shell_config_for(
        options.shell_to_configure, vexrc, environ)
    if not data:
        raise OtherShell("unknown shell: " + options.shell_to_configure)
    if hasattr(sys.stdout, 'buffer'):
        sys.stdout.buffer.write(data)
    else:
        sys.stdout.write(data)
    return 0


def get_cwd(options):
    """Discover what directory the command should run in.
    """
    if not options.cwd:
        return None
    if not os.path.exists(options.cwd):
        raise InvalidCwd(
            "can't --cwd to invalid path {0!r}".format(options.cwd))
    return options.cwd


def get_virtualenv_path(options, vexrc, environ):
    """Find a virtualenv path.
    """
    if options.path:
        ve_path = options.path
    else:
        ve_base = vexrc.get_ve_base(environ)
        if not ve_base:
            raise NoVirtualenvsDirectory(
                "could not figure out a virtualenvs directory. "
                "make sure $HOME is set, or $WORKON_HOME,"
                " or set virtualenvs=something in your .vexrc")
        # Using this requires get_ve_base to pass through nonexistent dirs
        if not os.path.exists(ve_base):
            raise NoVirtualenvsDirectory(
                "virtualenvs directory {0!r} not found.".format(ve_base))
        ve_name = options.rest.pop(0) if options.rest else ''
        if not ve_name:
            raise NoVirtualenvName(
                "could not find a virtualenv name in the command line."
            )

        # n.b.: if ve_name is absolute, ve_base is discarded by os.path.join,
        # and an absolute path will be accepted as first arg.
        # So we check if they gave an absolute path as ve_name.
        # But we don't want this error if $PWD == $WORKON_HOME,
        # in which case 'foo' is a valid relative path to virtualenv foo.
        ve_path = os.path.join(ve_base, ve_name)
        if ve_path == ve_name and os.path.basename(ve_name) != ve_name:
            raise InvalidVirtualenv(
                'To run in a virtualenv by its path, '
                'use "vex --path {0}"'.format(ve_path))

    ve_path = os.path.abspath(ve_path)
    if not os.path.exists(ve_path):
        raise InvalidVirtualenv("no virtualenv found at {0!r}.".format(ve_path))
    return ve_path


def get_command(options, vexrc, environ):
    """Get a command to run.

    :returns:
        a list of strings representing a command to be passed to Popen.
    """
    command = options.rest
    if not command:
        command = vexrc.get_shell(environ)
    if command and command[0].startswith('--'):
        raise InvalidCommand(
            "don't put flags like '%s' after the virtualenv name."
            % command[0])
    if not command:
        raise InvalidCommand("no command given")
    return command


def _main(environ, argv):
    """Logic for main(), with less direct system interaction.
    """
    options = get_options(argv)
    vexrc = get_vexrc(options, environ)
    if options.shell_to_configure:
        return handle_shell_config(options, vexrc, environ)
    cwd = get_cwd(options)
    try:
        ve_path = get_virtualenv_path(options, vexrc, environ)
    except NoVirtualenvName:
        options.print_help()
        raise
    command = get_command(options, vexrc, environ)
    env = get_environ(environ, vexrc['env'], ve_path)
    returncode = run(command, env=env, cwd=cwd)
    if returncode is None:
        raise InvalidCommand("command not found: {0!r}".format(command[0]))
    return returncode


def main():
    """The main command-line entry point, with system interactions.
    """
    argv = sys.argv[1:]
    returncode = 1
    try:
        returncode = _main(os.environ, argv)
    except NoVirtualenvName:
        sys.stderr.write("\n")
        sys.stderr.write("Error: no virtualenv name found in command line.\n")
    except InvalidArgument as error:
        if error.message:
            sys.stderr.write("Error: " + error.message + '\n')
    sys.exit(returncode)
